package CFG;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import AST.ASTCompoundStatement;
import AST.ASTExpressionStatement;
import AST.ASTIfStatement;
import AST.ASTNode;
import AST.ASTPrintStatement;
import AST.ASTReturnStatement;
import AST.ASTStatement;
import AST.ASTVisitor;
import AST.ASTWhileStatement;
import AST.ASTFunction;

/**
 * This class gives a "view" of the AST as an Control Flow Graph
 * with some of the ASTNodes as nodes.
 * 
 * @author Matthias Braun
 */
public class ASTCFG implements CFG<ASTCFGNode> {
	public ASTCFG(ASTFunction function) {
		start = createNode("Start");
		end = createNode("End");
		
		CFGBuilder cfgbuilder = new CFGBuilder();
		cfgbuilder.last = start;
		
		function.getStatement().accept(cfgbuilder);
		if (cfgbuilder.last != null)
			addEdge(cfgbuilder.last, end);
	}	
	
	public ASTCFGNode getStart() {
		return start;
	}
	
	public ASTCFGNode getEnd() {
		return end;
	}
	
	/** return successors of a node in the CFG */
	public List<ASTCFGNode> getSuccs(ASTCFGNode node) {
		List<ASTCFGNode> succs = this.succs.get(node);
		if (succs == null) {
			succs = new ArrayList<ASTCFGNode>();
			this.succs.put(node, succs);
		}
		return succs;
	}
	
	/** return predecessors of a node in the CFG */
	public List<ASTCFGNode> getPreds(ASTCFGNode node) {
		List<ASTCFGNode> preds = this.preds.get(node);
		if (preds == null) {
			preds = new ArrayList<ASTCFGNode>();
			this.preds.put(node, preds);
		}
		return preds;
	}
	
	
	/** add an edge to the CFG */
	private void addEdge(ASTCFGNode from, ASTCFGNode to) {
		if (from == null || to == null)
			return;
		
		getSuccs(from).add(to);
		getPreds(to).add(from);
	}
	
	/** Visitor that constructs the actual CFG from the AST */
	private class CFGBuilder extends ASTVisitor.StatementVisitor {
		public ASTCFGNode last;
		
		public void visitExpressionStatement(ASTExpressionStatement expressionStatement) {
			ASTCFGNode expr = createNode(expressionStatement.getExpression());
			addEdge(last, expr);
			last = expr;
		}

		public void visitIfStatement(ASTIfStatement if_statement) {
			ASTCFGNode ifnode = createNode(if_statement);
			addEdge(last, ifnode);
			/* create a magical statement behind the if */
			ASTCFGNode behind = createNode("");
			
			/* process then clause */
			last = ifnode;
			if_statement.getIfClause().accept(this);
			addEdge(last, behind);
			
			/* process else clause */
			last = ifnode;
			if_statement.getElseClause().accept(this);
			addEdge(last, behind);
			
			last = behind;								
		}

		public void visitPrintStatement(ASTPrintStatement ps) {
			ASTCFGNode cfgnode = createNode(ps);
			addEdge(last, cfgnode);
			last = cfgnode;
		}

		public void visitReturnStatement(ASTReturnStatement ret) {
			ASTCFGNode retNode = createNode(ret);
			addEdge(last, retNode);
			addEdge(retNode, end);
			last = null;
		}

		public void visitWhileStatement(ASTWhileStatement while_statement) {
			ASTCFGNode whileNode = createNode(while_statement);
			addEdge(last, whileNode);
			last = whileNode;
			/* build body */
			while_statement.getBody().accept(this);
			/* backedge */
			addEdge(last, whileNode);
			/* edge leaving the loop starts at loop header */
			last = whileNode;
		}

		public void visitCompoundStatement(ASTCompoundStatement compound) {
			for(ASTStatement statement : compound.getStatements()) {
				statement.accept(this);
			}
		}		
	}
	
	/** return node with id @p id */
	public ASTCFGNode nodeFromId(int id) {
		return nodes.get(id);
	}
	
	/** return biggest node id (+1) */
	public int maxNodeId() {
		return nodes.size();
	}
	
	private ASTCFGNode createNode(String name) {
		int id = nodes.size();
		ASTCFGNode node = new ASTCFGNode(name, id);
		nodes.add(node);
		
		return node;
	}
	
	private ASTCFGNode createNode(ASTNode astnode) {
		int id = nodes.size();
		ASTCFGNode node = new ASTCFGNode(astnode, id);
		nodes.add(node);
		
		return node;
	}
	
	/* "virtual" start and end node */
	private ArrayList<ASTCFGNode> nodes = new ArrayList<ASTCFGNode>();
	private ASTCFGNode start;
	private ASTCFGNode end;
	private Map<ASTCFGNode, List<ASTCFGNode>> preds = new HashMap<ASTCFGNode, List<ASTCFGNode>>();
	private Map<ASTCFGNode, List<ASTCFGNode>> succs = new HashMap<ASTCFGNode, List<ASTCFGNode>>();
}
