grammar Simple;

options {
	language = Java;
	k = 2;
}

@header {
package Parser;

import AST.ASTAssignExpression;
import AST.ASTCompareExpression;
import AST.ASTConstExpression;
import AST.ASTExpressionStatement;
import AST.ASTFunction;
import AST.ASTCompoundStatement;
import AST.ASTAddExpression;
import AST.ASTExpression;
import AST.ASTIfStatement;
import AST.ASTMulExpression;
import AST.ASTPrintStatement;
import AST.ASTReferenceExpression;
import AST.ASTReturnStatement;
import AST.ASTStatement;
import AST.ASTSymbol;
import AST.ASTWhileStatement;
import AST.SymbolTable;
import AST.ASTCompareExpression.Type;

}

@lexer::header {
package Parser;
}

@members {
	private final SymbolTable symbolTable = SymbolTable.getInstance();
	public boolean hadErrors = false;
  
	@Override
	public void emitErrorMessage(String message) {
		super.emitErrorMessage(message);
		hadErrors = true;
	}
}

public function returns [ASTFunction f]
	: stmts=statementList EOF
		{ $f = new ASTFunction(stmts); }
;

private statement returns [ASTStatement s]
	: c=compoundStatement {$s = c;} 
	| i=ifStatement {$s = i;}
	| p=printStatement {$s = p;}
	| e=expressionStatement {$s = e;}
	| w=whileStatement {$s = w;}
	| r=returnStatement {$s = r;}
;

private statementList returns [ASTCompoundStatement s]
	: { $s = new ASTCompoundStatement(); } 
	( stmt=statement
		{ $s.addStatement(stmt); }
	)*
;

private compoundStatement returns [ASTCompoundStatement s]
	: '{' slst=statementList '}' 
		{ $s = slst; }
;

private whileStatement returns [ASTStatement s]
	: 'while' '(' c=expression ')' st=compoundStatement
		{ $s = new ASTWhileStatement(c, st); }
;

private ifStatement returns [ASTIfStatement s]
	: 'if' '(' cond=expression ')' ifcl=statement
	(options {  k=1; }: 'else' elsecl=statement)? 
		{ $s = new ASTIfStatement(cond, ifcl, elsecl); }
;

private printStatement returns [ASTPrintStatement s]
	: 'print' '(' e=expression ')' ';'
		{ $s = new ASTPrintStatement(e); }
;

private expressionStatement returns [ASTStatement s]
	: e=expression ';'
		{ $s = new ASTExpressionStatement(e); }
;

private returnStatement returns [ASTReturnStatement s]
	: 'return' e=expression? ';' 
		{ $s = new ASTReturnStatement(e); }
;

private expression returns [ASTExpression exp]
	: ae=assignmentExpression
		{ $exp = ae; }
;

private assignmentExpression returns [ASTExpression exp]
	: var=identifier '=' right=assignmentExpression
		{ $exp = new ASTAssignExpression(var, right); }
	| expr=compareExpression
		{ $exp = expr; } 
;

private compareExpression returns [ASTExpression exp]
	: left=additiveExpression
		{ 
			$exp = left;
			ASTCompareExpression.Type compareType = Type.LESS;
		} 
	( ('<' { compareType = Type.LESS; }
	| '<=' { compareType = Type.LESSEQUAL; }
	| '>' { compareType = Type.GREATER; }
	| '>=' { compareType = Type.GREATEREQUAL; }
	| '==' { compareType = Type.EQUAL; }
	| '!=' { compareType = Type.NOTEQUAL; }) right=additiveExpression 
		{ $exp = new ASTCompareExpression($exp, right, compareType); }
	)*
;

private additiveExpression returns [ASTExpression exp]
	: left=multiplicativeExpression
		{ $exp = left; }
	(
		{ boolean is_plus = true; }
	('+'|'-' {is_plus = false; } ) right=multiplicativeExpression 
		{ $exp = new ASTAddExpression($exp, right, is_plus); }
	)*
;

private multiplicativeExpression returns [ASTExpression exp]
	: left=primaryExpression
		{ $exp = left; }
	( '*' right=primaryExpression 
		{ $exp = new ASTMulExpression($exp, right); }
	)*
;

private primaryExpression returns [ASTExpression exp]
	: i=INTEGER_LITERAL 
		{
			int val = Integer.parseInt(i.getText());
			$exp = new ASTConstExpression(val);
		}
	| id=identifier { $exp = new ASTReferenceExpression(id); }
	| '(' e=expression ')' { $exp = e; }
;

private identifier returns [ASTSymbol sym]
	: id=IDENTIFIER
		{ $sym = symbolTable.findOrInsert(id.getText()); }
;

/* Lexer */
COMMENT : '/*' (options {greedy=false;} : .)* '*/' { skip(); } ;
LINECOMMENT: '//' (options {greedy=false;} : .)* '\n' {skip();}  ;
fragment CHAR : ('a'..'z' | 'A'..'Z' | '_');
fragment DIGIT : ('0'..'9');
IDENTIFIER : CHAR (CHAR|DIGIT)*;
INTEGER_LITERAL : (DIGIT)+;
WHITESPACE : (' ' | '\t' | '\r' | '\n' )+
  { skip(); }
;